from collections import OrderedDict
from dvc_cc.version import VERSION
from dvc_cc.cli_modes import cli_modes

import os
import yaml
import requests
import keyring
from dvc.repo import Repo as DVCRepo
from git import Repo as GITRepo
from argparse import ArgumentParser
import datetime
from dvc_cc.hyperopt.variable import *
import subprocess
from pathlib import Path
from dvc_cc.bcolors import *

def get_main_git_directory_Path():
    gitrepo = GITRepo('.')
    git_path = gitrepo.common_dir.split('/.git')[0]
    return git_path


DESCRIPTION = 'With this command you can manage the hyperparameter that you already defined with '+bcolors.OKBLUE+'dvc-cc hyperopt new'+bcolors.ENDC+'.'

def main():
    parser = ArgumentParser(description=DESCRIPTION)
    parser.add_argument('name_of_variable', help='The name of variable to show. Use "all" to show all variables. If you use "all" --set and --set-type has no effect.', type=str)
    parser.add_argument('--set', help='Set a constant value for one parameter. If a constant value is set, it will not be queried when running "dvc-cc run". Use "None" to undo the constant value.', type=str, default=None)
    parser.add_argument('--set-type', help='If you want to reset the type of the variable you can use this parameter. This should be one of the following: float, int, file or one_of.', type=str, default=None)
    args = parser.parse_args()

    # go to the main git directory.
    #os.chdir(str(Path(get_main_git_directory_Path())))

    if not os.path.exists('dvc'):
        os.mkdir('dvc')

    if not os.path.exists(str(Path('dvc/.hyperopt'))):
        os.mkdir(str(Path('dvc/.hyperopt')))

    ######################
    # Read all Variables #
    ######################
    vc = VariableCache()
    if os.path.exists('dvc') and os.path.exists(str(Path('dvc/.hyperopt'))):
        list_of_hyperopt_files = [f for f in os.listdir(str(Path('dvc/.hyperopt'))) if f.endswith('.hyperopt')]
    else:
        list_of_hyperopt_files = []
    for f in list_of_hyperopt_files:
        f = str(Path('dvc/.hyperopt/' + f))
        vc.register_dvccc_file(f)

    if args.name_of_variable.upper() == 'ALL':
        print('%25s%8s%6s'%('Varname','type','value'))
        for v in vc.list_of_all_variables:
            print(v.__pretty_str__())
    else:
        v = Variable.search_varname_in_list(vc.list_of_all_variables, args.name_of_variable)
        if v is not None:
            if args.set_type is not None:
                v.set_type_of_variable(args.set_type)
            if args.set is not None:
                v.set_constant_value(args.set)
            if args.set_type is not None or args.set is not None:
                vc.update_dvccc_files()
            print(v.__pretty_str__())

        else:
            print(bcolors.FAIL+'Did not found the Variable: ' + str(args.name_of_variable)+ bcolors.ENDC)
            print("Please use one of the following variables:")

            print('%25s%8s%6s'%('Varname','type','value'))
            for v in vc.list_of_all_variables:
                print(v.__pretty_str__())
            exit(1)





















import random
from dvc_cc.hyperopt.variable import *
import re
import numpy as np
from dvc_cc.bcolors import *
from pathlib import Path

class HyperOptimizerBase:
    is_global = False
    allowed_types = []
    num_of_draws = 1
    variable = None
    name = None
    shortname = None

    def draw(self,):
        return None
    def set_settings(self,):
        return
    def help(self,):
        return ''

def read_value(text,typeofvalue, order=2):
    user_input = None
    while user_input == None:
        try:
            user_input = typeofvalue(input('\t'*order+text))
        except ValueError:
            print(bcolors.FAIL+'\t'*order+'Wrong datatype: Your input must be from the datatype ' + str(typeofvalue) + '.'+bcolors.ENDC)
            user_input = None
    return user_input

class HyperOptimizerGridSearch(HyperOptimizerBase):
    value = None
    minimum_value = None
    maximum_value = None
    allowed_types = ['float', 'ufloat', 'int', 'uint']      
    name = 'GridSearch'
    shortname = 'gs'
    is_global = False 
 
    def __init__(self, dtype, seed=None):
        self.dtype = dtype
        self.seed = seed

    def draw(self,):
        if self.dtype == 'int' or self.dtype == 'uint' or self.dtype == np.int or self.dtype == np.uint:
            return np.linspace(self.minimum_value,self.maximum_value,self.num_of_draws,dtype=int)
        else:
            return np.linspace(self.minimum_value,self.maximum_value,self.num_of_draws)

    def set_settings(self,):
        self.minimum_value = read_value('Min value: ', float)
        self.maximum_value = read_value('Max value: ', float)
        self.num_of_draws = read_value('Num of draws: ', int)

class HyperOptimizerRandomSearch(HyperOptimizerBase):
    value = None
    minimum_value = None
    maximum_value = None
    allowed_types = ['float', 'ufloat', 'int', 'uint']

    def __init__(self,dtype, seed=None):
        self.dtype = dtype
        self.seed = seed

    def draw(self,num_of_draws=None):
        if num_of_draws is not None:
            self.num_of_draws = num_of_draws
        if self.num_of_draws is None:
            r = np.random.beta(self.a,self.b)
        else:
            r = np.random.beta(self.a,self.b,self.num_of_draws)
        if self.dtype == 'int' or self.dtype == 'uint' or self.dtype == np.int or self.dtype == np.uint:
            r *= self.maximum_value - self.minimum_value + 1
        else:
            r *= self.maximum_value - self.minimum_value
        r += self.minimum_value
        if self.dtype == 'int' or self.dtype == 'uint' or self.dtype == np.int or self.dtype == np.uint:
            r = np.array(r,dtype=np.int)
        return r

    def set_settings(self,):
        self.minimum_value = read_value('Min value: ', float)
        self.maximum_value = read_value('Max value: ', float)
        self.a = read_value('Value A of the Beta-Distribution: ', float)
        self.b = read_value('Value B of the Beta-Distribution: ', float)
        
        if self.is_global == False:
            self.num_of_draws = read_value('How many draws do you want to make?: ', int)
        else:
            self.num_of_draws = None
        return

class HyperOptimizerRandomSearchLocal(HyperOptimizerRandomSearch):
    name = 'RandomSearch-Local'
    shortname = 'rsl'
    is_global = False

    def __init__(self, dtype, seed=None):
        super().__init__(dtype, seed)
class HyperOptimizerRandomSearchGlobal(HyperOptimizerRandomSearch):
    name = 'RandomSearch-Global'
    shortname = 'rs'
    is_global = True

    def __init__(self, dtype, seed=None):
        super().__init__(dtype, seed)
    

class Constant(HyperOptimizerBase):
    name = 'Constant'
    shortname = 'c'
    allowed_types = ['float', 'int', 'file']

    def set_data(self,data, dtype_of_variable):
        data = str(data).replace(' ', '').split(',')
        self.num_of_draws = len(data)
        self.value = data
        if dtype_of_variable == 'int':
            [int(val) for val in self.value]
        elif dtype_of_variable == 'int':
            [float(val) for val in self.value]
        elif dtype_of_variable.startswith('['):
            dtype_of_variable = dtype_of_variable[1:-1].split(',')
            tmp = np.array([val in dtype_of_variable or (val.isdigit() and int(val) < len(dtype_of_variable)) for val in self.value])
            if len(tmp[tmp==False]) > 0:
                raise  ValueError('You need to set one of the following types: '+ str(self.value))
        if len(np.unique(self.value)) != len(self.value):
            raise ValueError('It is not allowed to use the same value twice: ' + str(self.value))

    def draw(self,):
        return self.value

    def set_settings(self,):
        self.set_data(input('Set one or comma seperated values: '))
        return

class One_Of(HyperOptimizerBase):
    name = 'one_of'
    shortname = 'o'
    allowed_types = ['one_of']
    v = None

    def set_data(self,inputdata, vartype):
        vartypes = vartype[1:-1].split(',')
        inputdata = str(inputdata).replace(' ', '').split(',')
        data = []
        for d in inputdata:
            if d in vartypes:
                data.append(d)
            elif d.isdigit():
                data.append(vartypes[int(d)])
            else:
                raise ValueError('The user data did not match any of the parameter.')

        self.num_of_draws = len(data)
        if self.num_of_draws == 0:
            raise ValueError('The user data did not match any of the parameter.')
        self.value = data

    def draw(self,):
        return self.value

    def set_settings(self,vartype):
        self.set_data(input('Set one or comma seperated values: '), vartype)
        return

class RegexFileSearch(HyperOptimizerBase):
    name = 'FileSearch'
    shortname = 'fs'
    allowed_types = ['file']

    def __init__(self, dtype):
        super().__init__(dtype)

    def draw(self,):
        return self.matched_files

    def set_settings(self,):
        regex = None
        while regex is None:
            regex = input('Set regex value that describes the Files: ')

            matched_files = []
            for root, dirs, files in os.walk('', topdown=True):
                root = root[2:]

                for f in files:
                    if root == '':
                        if re.match(regex, f):
                            matched_files.append(f)
                    else:
                        f = str(Path(root + '/' + f))
                        if re.match(regex, f):
                            matched_files.append(f)

            if len(matched_files) == 0:
                print(bcolors.FAIL+'   Warning the regex does not match any file.'+bcolors.ENDC)
                regex = None
            else:
                print('Your regex match '+bcolors.OKGREEN+ str(len(matched_files)) + ' Files'+bcolors.ENDC+'. here are 5 random Examples: ')
                for s in range(5):
                    print('\t\t- ' + random.sample(matched_files, 1)[0])
                if not input('Do you want to use this regex? [y,n]: ').lower().startswith('y'):
                    regex = None
        self.num_of_draws = len(matched_files)
        self.matched_files = matched_files


def get_possible_hyperparameter(variabletype):
    if variabletype == 'file':
        return [RegexFileSearch]
    elif variabletype.startswith('[') and variabletype.endswith(']'):
        return []
    else:
        return [HyperOptimizerGridSearch, HyperOptimizerRandomSearchLocal, HyperOptimizerRandomSearchGlobal]



def select_hyperparameteroptimizer(user_input, possible_hyperparameter, dtype_of_variable):
    if user_input.startswith('--'):
        for h in possible_hyperparameter:
            if h.shortname == user_input[2:].lower():
                return h
        return None
    else:
        m = Constant()
        try:
            m.set_data(user_input, dtype_of_variable)
        except:
            return None
        return m


def combine_combinations(combinations, is_global):
    r = np.array([[0]])
    one_global_hyperopt_was_set=False

    for i in range(len(combinations)):
        shape = r.shape
        if not is_global[i] or not one_global_hyperopt_was_set:
            r = np.tile(r, len(combinations[i])).reshape(-1, shape[1])
            next_draw = np.tile(combinations[i], shape[0]).reshape(-1, 1)
            if is_global[i]:
                one_global_hyperopt_was_set = True
        else:
            next_draw = np.tile(combinations[i], shape[0]//len(combinations[i])).reshape(-1, 1)


        r = np.append(r, next_draw , axis=1)

    return r[:,1:]

def create_hyperopt_variables(vc): # vc == VariableCache

    num_of_global_draws = None

    for v in vc.list_of_all_variables:
        print()
        if v.varvalue is None:
            hyper = get_possible_hyperparameter(v.vartype)
            selected_hyper = None
            while selected_hyper is None:
                print('Specifie the variable ' + v.varname)
                print('\tSet one or multiple COMMA exitsebarated values directly.')
                if len(hyper) > 0:
                    print('\tOr do hyperoptimization with one of the following options:')
                    for h in hyper:
                        print('\t\t- ' + h.name + ' (with '+bcolors.OKGREEN+'--' + h.shortname + bcolors.ENDC +')')

                user_input = input('\t'+ str(v.varname) + ' (' + bcolors.OKGREEN+str(v.vartype) + bcolors.ENDC+'): ')

                if v.vartype.startswith('[') and v.vartype.endswith(']'):
                    try:
                        selected_hyper = One_Of()
                        selected_hyper.set_data(user_input, v.vartype)
                    except:
                        print(bcolors.FAIL+'\t\tError: Your input did not match one of the possible values.'+bcolors.ENDC)
                        print('\t\tPlease use one of the values ' + bcolors.OKGREEN+str(v.vartype) + bcolors.ENDC + ' or use an index.')
                        selected_hyper = None
                else:
                    selected_hyper = select_hyperparameteroptimizer(user_input, hyper, v.vartype)
                if selected_hyper is None:
                    print(bcolors.FAIL+'\t\tWarning: did not understand which Hyperoptimizer you want to use.'+bcolors.ENDC)
            if type(selected_hyper) is not Constant and type(selected_hyper) is not One_Of:
                selected_hyper = selected_hyper(v.vartype)
                selected_hyper.set_settings()


            if selected_hyper.is_global:
                if num_of_global_draws is None:
                    num_of_global_draws = read_value('\tHow many global draws do you want to make?: ', int)
                selected_hyper.num_of_draws = num_of_global_draws

            v.hyperoptimizer = selected_hyper

        else:
            c = Constant()
            c.set_data(v.varvalue, v.vartype)
            v.hyperoptimizer = c

    # create a product over all variants
    is_global = [v.hyperoptimizer.is_global for v in vc.list_of_all_variables]
    draws = [v.hyperoptimizer.draw() for v in vc.list_of_all_variables]

    return combine_combinations(draws, is_global)






































#!/usr/bin/env python3
import sys
import os
from subprocess import check_output
import subprocess
from argparse import ArgumentParser
import keyring
import requests
import yaml
from dvc_cc.bcolors import *
from pathlib import Path
from git import Repo as GITRepo
import json
import time
import pexpect

DESCRIPTION = 'This script saves the SSHFS connection that was created with this script and can reconnect to this ' \
              'source. You can also use the "--keyring-service str_name" parameter.'


def get_main_git_directory_Path():
    gitrepo = GITRepo('.')
    git_path = gitrepo.common_dir.split('/.git')[0]
    return git_path

def get_mount_values_for_a_direcotry(path):
    mount = [m.split(' ') for m in check_output(["mount"]).decode("utf8").split('\n')]
    for m in mount:
        if len(m) == 6 and (m[2] == path or m[2] == os.getcwd() + '/' + path):
            username = m[0].split('@')[0]
            servername = m[0].split('@')[1].split(':')[0]
            path = m[0].split('@')[1].split(':')[1]
            return username, servername, path
    return None, None, None

def reconnect(keyring_service='red'):
    project_dir = get_main_git_directory_Path()
    path_to_sshfs_json = str(Path(os.path.join(project_dir, '.dvc_cc/sshfs.json')))

    if os.path.exists(path_to_sshfs_json):
        with open(path_to_sshfs_json, "r") as jsonFile:
            data = json.load(jsonFile)

        for key in data.keys():
            new_sshfs_connection([data[key]["username"] + '@' + data[key]["server"] + ':' + data[key]["remote_path"], key])

def unmount():
    project_dir = get_main_git_directory_Path()
    path_to_sshfs_json = str(Path(os.path.join(project_dir, '.dvc_cc/sshfs.json')))

    if os.path.exists(path_to_sshfs_json):
        with open(path_to_sshfs_json, "r") as jsonFile:
            data = json.load(jsonFile)

        for key in data.keys():
            if get_mount_values_for_a_direcotry(key)[0] is not None:
                print(check_output(['fusermount', '-u', key]).decode("utf8"))
                time.sleep(1)

def new_sshfs_connection(sshfs_parameters, keyring_service = 'red'):
    project_dir = get_main_git_directory_Path()

    dest = os.path.realpath(os.path.expanduser(sshfs_parameters[-1]))
    dest_rel = dest[len(project_dir) + 1:]

    # Prüfe ob Ordner existiert und lege ihn gegebenfalls neu an
    if not os.path.exists(dest):
        print(dest, os.path.exists(dest))
        os.makedirs(dest)
    # Prüfen ob eine SSHFS-Verbindung für das Ziel bereits vorliegt
    elif get_mount_values_for_a_direcotry(dest)[0] is not None:
        # Wenn Ja
        # Unmount aktuelle Verbindung
        print('Unmount last connection!')
        print(check_output(['fusermount', '-u', dest]).decode("utf8"))

    # Schreibe den Ordner in die .gitignore!
    f= open(str(Path(os.path.join(project_dir, '.gitignore'))),"a+")
    f.write('\n'+dest_rel)
    f.close()

    # call the sshfs command with the password from keyring if possible.
    username = None
    pw_keyring = None
    username_keyring = None
    if sshfs_parameters[-2].find('@') > 0:
        username, server = sshfs_parameters[-2].split('@')
        server = server.split(':')[0]
        server = server.replace('.','_').replace('-','_')

        pw_keyring = server + '_password'
        username_keyring = server + '_username'

        pw_keyring = keyring.get_password(keyring_service, pw_keyring)
        username_keyring = keyring.get_password(keyring_service, username_keyring)

    if pw_keyring is not None and username_keyring is not None and username == username_keyring:
        # see here https://stackoverflow.com/questions/28823639/how-to-automatically-input-ssh-private-key-passphrase-with-pexpect
        bash = pexpect.spawn('bash', echo=True)
        bash.sendline('echo READY')
        bash.expect_exact('READY')
        bash.sendline('sshfs ' + ' '.join(sshfs_parameters[:-1]) + ' ' + dest)
        bash.expect('password')
        bash.sendline(pw_keyring)
        bash.sendline('echo COMPLETE')
        bash.expect_exact('COMPLETE')
        bash.sendline('exit')
        bash.expect_exact(pexpect.EOF)

    else:
        print(subprocess.call(['sshfs'] + sshfs_parameters[:-1] + [dest]))
    # Finden der wichtigen Parameter
    data_username, data_server, data_path = get_mount_values_for_a_direcotry(dest)
    if data_server is None or data_server == 'null':
        raise ValueError('The connection is lost. Maybe you do not have the rights?')

    # Bei gelingen: Erstelle einen neuen Eintrag oder aktualliesiere den alten Eintrag
    path_to_sshfs_json = str(Path(os.path.join(project_dir, '.dvc_cc/sshfs.json')))
    if os.path.exists(path_to_sshfs_json):
        with open(path_to_sshfs_json, "r") as jsonFile:
            data = json.load(jsonFile)
    else:
        data = {}

    new_sshfs_connection = {
        dest_rel: {
            "server": data_server,
            "remote_path": data_path,
            "username": data_username}}

    data.update(new_sshfs_connection)



    with open(path_to_sshfs_json, "w") as jsonFile:
        json.dump(data, jsonFile, indent=2, sort_keys=True)


def main():

    argv_tmp = sys.argv[1:]
    argv = []
    found_keyring_service = False
    keyring_service = 'red'
    for i in argv_tmp:
        if i == '--keyring-service':
            found_keyring_service = True
        elif found_keyring_service:
            keyring_service = i
            found_keyring_service = False
        else:
            argv.append(i)

    if '-h' in argv or '--help' in argv or len(argv) == 0:
        print(DESCRIPTION)
        print()
    elif len(argv) == 1 and 'reconnect'[:len(argv[0])] == argv[0]:
        print('Reconnect ALL SSHFS connections')
        reconnect(keyring_service)
    elif len(argv) == 1 and 'unmount'[:len(argv[0])] == argv[0]:
        print('UNMOUNT ALL SSHFS connections')
        unmount()
    elif len(argv) == 1:
        print('NEW SSHFS connection to data folder')
        new_sshfs_connection(argv + ['data'], keyring_service)
    else:
        print('NEW SSHFS connection')
        new_sshfs_connection(argv, keyring_service)

